import ch.bildspur.realsense.*;

import java.nio.*;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2ES2;

// virtual camera の設定
// virtual世界での距離単位1 = 実世界の1mmに相当

// 視点
final float EYE_X = 0;
final float EYE_Y = 1000;
final float EYE_Z = 2000;

// 注視点
final float TARGET_X = 0;
final float TARGET_Y = 1000;
final float TARGET_Z = 0;

// 上方向
final float UP_X = 0;
final float UP_Y = 1;
final float UP_Z = 0;

// 40.0はIntel RealSenseのデプスカメラの縦方向画角
final float FOVY = 40.0 * PI / 180.0;
final float NEAR = 100;
final float FAR = 10000;

final int CAMERA_WIDTH = 640;
final int CAMERA_HEIGHT = 480;
final int CAMERA_FPS = 30;

final int IMAGE_WIDTH = CAMERA_WIDTH;
final int IMAGE_HEIGHT = CAMERA_HEIGHT;

final int WINDOW_WIDTH = IMAGE_WIDTH * 2;
final int WINDOW_HEIGHT = IMAGE_HEIGHT;

final int DEPTH_MIN = 0;
final int DEPTH_MAX = 5000;

RealSenseCamera camera;

PGraphics pg;
IntBuffer pgColor;
FloatBuffer pgDepth;

PImage realColorImage;
int[] realDepthBuffer;

PImage virtualColorImage;
int[] virtualDepthBuffer;

PImage colorImage;
int[] depthBuffer;

void settings(){
  size(WINDOW_WIDTH, WINDOW_HEIGHT, P3D);
}

void setup()
{
  camera = new RealSenseCamera(this);
  camera.start(CAMERA_WIDTH, CAMERA_HEIGHT, CAMERA_FPS, true, true);
  
  pg = createGraphics(IMAGE_WIDTH, IMAGE_HEIGHT, P3D);
  pg.noSmooth();
  
  pgColor = ByteBuffer.allocateDirect((IMAGE_WIDTH * IMAGE_HEIGHT) << 2).order(ByteOrder.nativeOrder()).asIntBuffer();
  pgDepth = ByteBuffer.allocateDirect((IMAGE_WIDTH * IMAGE_HEIGHT) << 2).order(ByteOrder.nativeOrder()).asFloatBuffer();

  realDepthBuffer = new int[IMAGE_WIDTH * IMAGE_HEIGHT];
  virtualDepthBuffer = new int[IMAGE_WIDTH * IMAGE_HEIGHT];
  depthBuffer = new int[IMAGE_WIDTH * IMAGE_HEIGHT];
}

void draw()
{
  // read frames
  camera.readFrames();
  
  // get color
  realColorImage = camera.getColorImage();
  
  // get depth
  for (int y = 0; y < CAMERA_HEIGHT; y++) {
    for (int x = 0; x < CAMERA_WIDTH; x++) {
      realDepthBuffer[x + y * CAMERA_WIDTH] = camera.getDepth(x, y);
    }
  }  
  
  pg.beginDraw();
  pg.background(0, 128, 192);

  pg.camera(EYE_X, EYE_Y, EYE_Z, TARGET_X, TARGET_Y, TARGET_Z, UP_X, UP_Y, UP_Z);
  pg.perspective(FOVY, pg.width / float(pg.height), NEAR, FAR);
  pg.lights();
  
  pg.translate(0, 1000, 1000);
  pg.pushMatrix();
  float theta = TWO_PI * millis() / 5000;
  pg.rotateY(theta);
  pg.translate(500, 0, 0);
  pg.noStroke();
  pg.fill(0, 192, 0);
  pg.sphere(100);
  pg.popMatrix();  

  pg.endDraw();

  PGraphics3D pg3d = (PGraphics3D) pg;
  FrameBuffer fb = pg3d.getFrameBuffer();
  
  PJOGL pgl = (PJOGL) beginPGL();
  GL2ES2 gl = pgl.gl.getGL2ES2();
  
  gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, fb.glFbo);
  gl.glReadPixels(0, 0, IMAGE_WIDTH, IMAGE_HEIGHT, GL.GL_BGRA, GL.GL_UNSIGNED_BYTE, pgColor);
  gl.glReadPixels(0, 0, IMAGE_WIDTH, IMAGE_HEIGHT, PGL.DEPTH_COMPONENT, GL.GL_FLOAT, pgDepth);

  endPGL();
  
  // get color
  virtualColorImage = createImage(IMAGE_WIDTH, IMAGE_HEIGHT, RGB);
  virtualColorImage.loadPixels();
  for (int i = 0; i < virtualColorImage.pixels.length; i++) {
    virtualColorImage.pixels[i] = pgColor.get(i);
  }
  virtualColorImage.updatePixels();
  
  // get depth
  for (int y = 0; y < IMAGE_HEIGHT; y++) {
    for (int x = 0; x < IMAGE_WIDTH; x++) {
      float d = pgDepth.get(x + y * IMAGE_WIDTH);
      d = 2 * d - 1;
      float z = 2.0 * NEAR * FAR / (FAR + NEAR - d * (FAR - NEAR));
      virtualDepthBuffer[x + y * IMAGE_WIDTH] = int(z);
    }
  } 

  // ここまででrealとvirtualの両方の情報が揃った
  // 合成する．
  
  PImage colorImage = createImage(CAMERA_WIDTH, CAMERA_HEIGHT, RGB);
  PImage depthImage = createImage(CAMERA_WIDTH, CAMERA_HEIGHT, RGB);
  
  colorImage.loadPixels();
  depthImage.loadPixels();
  for (int y = 0; y < CAMERA_HEIGHT; y++) {
    for (int x = 0; x < CAMERA_WIDTH; x++) {
      int rd = realDepthBuffer[x + y * CAMERA_WIDTH]; // 実空間の距離
      int vd = virtualDepthBuffer[x + y * CAMERA_WIDTH]; // virtual空間の距離
      int c; // 描画する色
      int d; // 距離画像として描画する色
      if (false /* ここに適当な条件を記述する */) { // 実空間を描画する
        int c1 = int((1.0 - (rd - DEPTH_MIN) / float(DEPTH_MAX - DEPTH_MIN)) * 255);
        c = color(c1, c1, c1);
        d = color(c1, 0, 0);
      } else { // virtual空間を描画する
        int c1 = int((1.0 - (vd - DEPTH_MIN) / float(DEPTH_MAX - DEPTH_MIN)) * 255);
        c = virtualColorImage.get(x, y);
        d = color(c1, c1, c1);
      }
      colorImage.set(x, y, c);
      depthImage.set(x, y, d);
    }
  }
  colorImage.updatePixels();
  depthImage.updatePixels();

  image(colorImage, 0, 0);
  image(depthImage, CAMERA_WIDTH, 0);
}
